﻿Imports System.Collections.ObjectModel
Imports System.ComponentModel
Imports System.IO
Imports System.Text
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media

Namespace MiddlemarchReader
	Public Class PageProvider
		Implements IPageProvider
		Private Shared ReadOnly blackBrush As Brush = New SolidColorBrush(Colors.Black)

		Private paragraphs As New List(Of String)()
		Private pageSize As Size
		Private fontFamily As FontFamily
		Private fontSize As Double

		Public Property Book() As BookInfo

		Protected Sub PreprocessBook(ByVal stream As Stream)
			' Occurs every time program is run
			GenerateParagraphs(stream)

			' Occurs only first time program is run
			If Me.Book.Chapters.Count = 0 Then
				GenerateChapters()
			End If
		End Sub

		Private Sub GenerateParagraphs(ByVal stream As Stream)
			Using reader As New StreamReader(stream)
				Dim line As String
				Dim paragraph As New StringBuilder()

				line = reader.ReadLine()
				Do While Nothing IsNot line
					' Check for a possible paragraph termination
					If line.Length = 0 OrElse line.Chars(0) = " "c Then
						If paragraph.Length > 0 Then
							paragraphs.Add(paragraph.ToString())
							paragraph.Remove(0, paragraph.Length)
						End If

						' Add the line anyway
						paragraphs.Add(line)
					Else
						' Insert a space between appended lines assembled into paragraph
						If paragraph.Length > 0 Then
							paragraph.Append(" ")
						End If

						paragraph.Append(line)
					End If
					line = reader.ReadLine()
				Loop

				' Add the last paragraph if there's one in progress
				If paragraph.Length > 0 Then
					paragraphs.Add(paragraph.ToString())
				End If
			End Using
		End Sub

		Private Sub GenerateChapters()
			' Set so the beginning becomes the first chapter
			Dim consecutiveBlankLineCount As Integer = 10
			Dim accumulatedCharacterCount As Integer = 0

			For i As Integer = 0 To paragraphs.Count - 1
				If paragraphs(i).Length = 0 Then
					consecutiveBlankLineCount += 1
				Else
					If consecutiveBlankLineCount > 2 Then
						' Create a new chapter
						Dim pageInfo As PageInfo = New PageInfo With {.ParagraphIndex = i, .CharacterIndex = 0, .AccumulatedCharacterCount = accumulatedCharacterCount}

						Dim chapterInfo As ChapterInfo = New ChapterInfo With {.Title = paragraphs(i)}
						chapterInfo.Pages.Add(pageInfo)
						Me.Book.Chapters.Add(chapterInfo)
					End If
					consecutiveBlankLineCount = 0
					accumulatedCharacterCount += paragraphs(i).Length
				End If
			Next i

			' Flag the last chapter
			Me.Book.Chapters(Me.Book.Chapters.Count - 1).IsLastChapter = True
			Me.Book.CharacterCount = accumulatedCharacterCount
		End Sub

        Public Sub SetPageSize(ByVal size As Size) Implements IPageProvider.SetPageSize
            If pageSize <> size Then
                pageSize = size

                ' TODO: If the page size can really change,
                '      pagination information needs to be invalidated
            End If
        End Sub

		Public Sub SetFont(ByVal fontFamily As FontFamily, ByVal fontSize As Double) Implements IPageProvider.SetFont
			Me.fontFamily = fontFamily
			Me.fontSize = fontSize

			' TODO: If the font and font size can really change,
			'      pagination information needs to be invalidated
		End Sub


		Public Function QueryLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.QueryLastPage
			Return GetLastPage(chapterIndex, pageIndex, False)
		End Function

		Public Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetLastPage
			Return GetLastPage(chapterIndex, pageIndex, True)
		End Function

		Private Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer, ByVal paginateIfNecessary As Boolean) As FrameworkElement
			If pageSize.IsEmpty Then
				pageIndex = -1
				Return Nothing
			End If

			If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
				pageIndex = -1
				Return Nothing
			End If

			Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

			If chapter.Pages(chapter.Pages.Count - 1).IsLastPage Then
				pageIndex = chapter.Pages.Count - 1
				Dim nextPageInfo As PageInfo = If(chapter.IsLastChapter, Nothing, Me.Book.Chapters(chapterIndex + 1).Pages(0))
				Return GetPaginatedPage(chapter.Pages(chapter.Pages.Count - 1), nextPageInfo)
			End If

			If Not paginateIfNecessary Then
				pageIndex = -1
				Return Nothing
			End If

			Dim element As FrameworkElement = Nothing

			Do While Not chapter.Pages(chapter.Pages.Count - 1).IsLastPage
				element = GetPage(chapterIndex, chapter.Pages.Count - 1)
			Loop

			pageIndex = chapter.Pages.Count - 1
			Return element
		End Function

		Public Function GetPage(ByVal chapterIndex As Integer, ByVal pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetPage
			If pageSize.IsEmpty Then
				Return Nothing
			End If

			If pageIndex < 0 Then
				Return Nothing
			End If

			If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
				Return Nothing
			End If

			Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

			' Check if the last page of the paragraph has already been established, 
			'  and the requested page is greater than that last page
			If chapter.Pages(chapter.Pages.Count - 1).IsLastPage AndAlso pageIndex > chapter.Pages.Count - 1 Then
				Return Nothing
			End If

			' If the last page of a chapter is requested, and it's already available, get it
			If chapter.Pages(pageIndex).IsLastPage Then
				' Last chapter in book and hence last page in book
				If chapterIndex = Me.Book.Chapters.Count - 1 Then
					Return GetPaginatedPage(chapter.Pages(pageIndex), Nothing)
				End If

				' The last page of a chapter but not the last chapter
				Return GetPaginatedPage(chapter.Pages(pageIndex), Me.Book.Chapters(chapterIndex + 1).Pages(0))
			End If

			' Otherewise, if the page has already been paginated, get it
			If pageIndex + 1 < chapter.Pages.Count Then
				Return GetPaginatedPage(chapter.Pages(pageIndex), chapter.Pages(pageIndex + 1))
			End If

			If pageIndex <> chapter.Pages.Count - 1 Then
				Throw New Exception(String.Format("GetPage can only paginate one page per call: pageIndex = {0} while chapter.Pages.Count = {1}", pageIndex, chapter.Pages.Count))
			End If

			Dim stackPanel As New StackPanel()
			Dim pageInfo As PageInfo = chapter.Pages(pageIndex)
			Dim paragraphIndex As Integer = pageInfo.ParagraphIndex
			Dim characterIndex As Integer = pageInfo.CharacterIndex
			Dim accumulatedCharacterCount As Integer = pageInfo.AccumulatedCharacterCount

			' Get the paragraph index of the next chapter, if it exists
			Dim nextChapterParagraphIndex As Integer = paragraphs.Count

			If chapterIndex + 1 < Me.Book.Chapters.Count Then
				nextChapterParagraphIndex = Me.Book.Chapters(chapterIndex + 1).Pages(0).ParagraphIndex
			End If

			Do While paragraphIndex < paragraphs.Count AndAlso paragraphIndex < nextChapterParagraphIndex
				' Skip if a blank line is the first paragraph on a page
				If stackPanel.Children.Count = 0 AndAlso characterIndex = 0 AndAlso paragraphs(paragraphIndex).Length = 0 Then
					paragraphIndex += 1
					Continue Do
				End If

				Dim txtblk As TextBlock = CreateTextBlock(paragraphs(paragraphIndex).Substring(characterIndex))
				stackPanel.Children.Add(txtblk)
				stackPanel.Measure(New Size(pageSize.Width, Double.PositiveInfinity))

				' Check if the StackPanel fits in the available height
				If stackPanel.DesiredSize.Height > pageSize.Height Then
					' Strip words off the end until it fits
					Do
						Dim index As Integer = txtblk.Text.LastIndexOf(" "c)

						If index = -1 Then
							index = 0
						End If

						txtblk.Text = txtblk.Text.Substring(0, index)
						stackPanel.Measure(New Size(pageSize.Width, Double.PositiveInfinity))

						If index = 0 Then
							Exit Do
						End If
					Loop While stackPanel.DesiredSize.Height > pageSize.Height

					characterIndex += txtblk.Text.Length
					accumulatedCharacterCount += txtblk.Text.Length

					' Skip over the space
					If txtblk.Text.Length > 0 Then
						characterIndex += 1
					End If

					Exit Do
				End If
				accumulatedCharacterCount += txtblk.Text.Length
				paragraphIndex += 1
				characterIndex = 0
			Loop

			' Skip over blank paragraphs. We don't want these at the top
			'   of the next page!
			Do While paragraphIndex < nextChapterParagraphIndex
				If paragraphs(paragraphIndex).Length = 0 Then
					paragraphIndex += 1
				Else
					Exit Do
				End If
			Loop

			pageInfo.IsLastPage = paragraphIndex = nextChapterParagraphIndex

			' Create a PageInfo for the next page only if it's not the
			'  last page in the chapter
			If Not pageInfo.IsLastPage Then
				Dim nextPageInfo As PageInfo = New PageInfo With {.ParagraphIndex = paragraphIndex, .CharacterIndex = characterIndex, .AccumulatedCharacterCount = accumulatedCharacterCount}
				chapter.Pages.Add(nextPageInfo)
			End If

			Return stackPanel
		End Function

		' Creates a page from one PageInfo to another.
		' NOTE: pageInfo2 is null if this is the last page of the book.
		Private Function GetPaginatedPage(ByVal pageInfo1 As PageInfo, ByVal pageInfo2 As PageInfo) As FrameworkElement
			Dim pnl As Panel = New StackPanel()

			' This is the case where a page has only part of a single paragraph
			If pageInfo2 IsNot Nothing AndAlso pageInfo1.ParagraphIndex = pageInfo2.ParagraphIndex Then
				Dim text As String = paragraphs(pageInfo1.ParagraphIndex).Substring(pageInfo1.CharacterIndex, pageInfo2.CharacterIndex - pageInfo1.CharacterIndex)
				pnl.Children.Add(CreateTextBlock(text))
			Else
				' Add first (possibly) partial paragraph
				Dim text As String = paragraphs(pageInfo1.ParagraphIndex).Substring(pageInfo1.CharacterIndex)
				pnl.Children.Add(CreateTextBlock(text))

				' Set paragraphIndex2 based on  pageInfo2 possibly being null
				Dim paragraphIndex2 As Integer = If(pageInfo2 IsNot Nothing, pageInfo2.ParagraphIndex, paragraphs.Count)

				' Add full paragraphs
				For paragraphIndex As Integer = pageInfo1.ParagraphIndex + 1 To paragraphIndex2 - 1
					text = paragraphs(paragraphIndex)
					pnl.Children.Add(CreateTextBlock(text))
				Next paragraphIndex

				' Add partial paragraph at end
				If pageInfo2 IsNot Nothing Then
					text = paragraphs(pageInfo2.ParagraphIndex).Substring(0, pageInfo2.CharacterIndex)
					pnl.Children.Add(CreateTextBlock(text))
				End If
			End If
			Return pnl
		End Function

		Private Function CreateTextBlock(ByVal text As String) As TextBlock
			Dim isBlankLine As Boolean = String.IsNullOrEmpty(text)

			Return New TextBlock With {.FontFamily = Me.fontFamily, .FontSize = Me.fontSize, .Text = If(isBlankLine, " ", text), .TextWrapping = TextWrapping.Wrap, .Foreground = blackBrush, .LineStackingStrategy = If(isBlankLine, LineStackingStrategy.BlockLineHeight, LineStackingStrategy.MaxHeight), .LineHeight = fontSize / (If(isBlankLine, 2, 1))}
		End Function
	End Class
End Namespace

